from __future__ import annotations

from os import fsdecode, path

from awkward._dispatch import high_level_function

__all__ = ("to_parquet_dataset",)


@high_level_function()
def to_parquet_dataset(
    directory,
    filenames=None,
    storage_options=None,
):
    """
    Args:
        directory (str or Path): A directory in which to write `_common_metadata`
            and `_metadata`, making the directory of Parquet files into a dataset.
        filenames (None or list of str or Path): If None, the `directory` will be
            recursively searched for files ending in `filename_extension` and
            sorted lexicographically. Otherwise, this explicit list of files is
            taken and row-groups are concatenated in its given order. If any
            filenames are relative, they are interpreted relative to `directory`.
        filename_extension (str): Filename extension (including `.`) to use to
            search for files recursively. Ignored if `filenames` is None.

    Creates a `_common_metadata` and a `_metadata` in a directory of Parquet files.

        >>> ak.to_parquet(array1, "/directory/arr1.parquet", parquet_compliant_nested=True)
        >>> ak.to_parquet(array2, "/directory/arr2.parquet", parquet_compliant_nested=True)
        >>> ak.to_parquet_dataset("/directory")

    The `_common_metadata` contains the schema that all files share. (If the files
    have different schemas, this function raises an exception.)

    The `_metadata` contains row-group metadata used to seek to specific row-groups
    within the multi-file dataset.
    """

    return _impl(directory, filenames, storage_options)


def _impl(directory, filenames, storage_options):
    # Implementation
    import awkward._connect.pyarrow

    pyarrow_parquet = awkward._connect.pyarrow.import_pyarrow_parquet(
        "ak.to_parquet_dataset"
    )
    import fsspec.parquet

    try:
        directory = fsdecode(directory)
    except TypeError:
        raise TypeError(
            f"'directory' argument of 'ak.to_parquet_dataset' must be a path-like, not {type(directory).__name__} ('array' argument is first; 'destination' second)"
        ) from None
    fs, destination = fsspec.core.url_to_fs(directory, **(storage_options or {}))
    if not fs.isdir(destination):
        raise ValueError(f"{destination!r} is not a directory" + {__file__})

    filepaths = get_filepaths(filenames, fs, destination)

    if len(filepaths) == 0:
        raise ValueError(f"no *.parquet or *.parq matches for path {destination!r}")

    schema = None
    metadata_collector = []
    for filepath in filepaths:
        with fs.open(filepath, mode="rb") as f:
            if schema is None:
                schema = pyarrow_parquet.ParquetFile(f).schema_arrow
                first_filepath = filepath
            elif not schema.equals(pyarrow_parquet.ParquetFile(f).schema_arrow):
                raise ValueError(
                    f"schema in {filepath!r} differs from the first schema (in {first_filepath!r})"
                )
            metadata_collector.append(pyarrow_parquet.ParquetFile(f).metadata)
            metadata_collector[-1].set_file_path(filepath)

    _common_metadata_path = path.join(destination, "_common_metadata")
    pyarrow_parquet.write_metadata(schema, _common_metadata_path, filesystem=fs)

    _metadata_path = path.join(destination, "_metadata")
    pyarrow_parquet.write_metadata(
        schema, _metadata_path, metadata_collector=metadata_collector, filesystem=fs
    )
    return _common_metadata_path, _metadata_path


def get_filepaths(filenames, fs, destination):
    filepaths = []
    if filenames is not None:
        if isinstance(filenames, str):
            for f in fs.glob(path.join(destination, filenames)):
                if f.endswith((".parq", ".parquet")):
                    filepaths.append(f)
        else:
            for filename in filenames:
                for f in fs.glob(path.join(destination, filename)):
                    if f.endswith((".parq", ".parquet")):
                        filepaths.append(f)
    else:
        for f, fdata in fs.find(destination, detail=True).items():
            if f.endswith((".parq", ".parquet")):
                if fdata["type"] == "file":
                    filepaths.append(f)
    return filepaths



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_numpy_array(forget_length, recursive):
    matrix = np.arange(64).reshape(8, -1)
    layout = ak.contents.NumpyArray(matrix[:, 0])
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_empty_array(forget_length, recursive):
    layout = ak.contents.EmptyArray()
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_indexed_option_array(forget_length, recursive):
    index = ak.index.Index64(np.r_[0, -1, 2, -1, 4])
    content = ak.contents.NumpyArray(np.arange(8))
    layout = ak.contents.IndexedOptionArray(index, content)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_indexed_array(forget_length, recursive):
    index = ak.index.Index64(np.array([0, 1, 2, 3, 6, 7, 8]))
    content = ak.contents.NumpyArray(np.arange(10))
    layout = ak.contents.IndexedArray(index, content)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_list_array(forget_length, recursive):
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.index.Index64(np.array([3, 3, 5, 6, 9]))
    layout = ak.contents.ListArray(starts, stops, content)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_list_offset_array(forget_length, recursive):
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    layout = ak.contents.ListOffsetArray(offsets, content)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_unmasked_array(forget_length, recursive):
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    layout = ak.contents.UnmaskedArray(content)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_union_array(forget_length, recursive):
    a = ak.contents.NumpyArray(np.arange(4))
    b = ak.contents.NumpyArray(np.arange(4) + 4)
    c = ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(12)), 3)
    layout = ak.contents.UnionArray.simplified(
        ak.index.Index8([1, 1, 2, 2, 0, 0]),
        ak.index.Index64([0, 1, 0, 1, 0, 1]),
        [a, b, c],
    )
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_record_array(forget_length, recursive):
    a = ak.contents.NumpyArray(np.arange(10))
    b = ak.contents.NumpyArray(np.arange(10) * 2 + 4)
    layout = ak.contents.RecordArray([a, b], None, 5)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_regular_array(forget_length, recursive):
    content = ak.contents.NumpyArray(np.arange(10))
    layout = ak.contents.RegularArray(content, 3)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_bit_masked_array(forget_length, recursive):
    mask = ak.index.IndexU8(np.array([0b10101010]))
    content = ak.contents.NumpyArray(np.arange(16))
    layout = ak.contents.BitMaskedArray(mask, content, False, 8, False)
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )


@pytest.mark.parametrize("forget_length", [True, False])
@pytest.mark.parametrize("recursive", [True, False])
def test_byte_masked_array(forget_length, recursive):
    mask = ak.index.Index8(np.array([1, 0, 1, 0, 1, 0, 1, 0]))
    content = ak.contents.NumpyArray(np.arange(16))
    layout = ak.contents.ByteMaskedArray(
        mask,
        content,
        False,
    )
    assert (
        layout.to_typetracer(forget_length).to_packed(recursive).form
        == layout.to_packed(recursive).form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
#
from __future__ import annotations

import os

import pytest

safetensors = pytest.importorskip("safetensors")


def test_roundtrip():
    import awkward as ak

    array = ak.Array([[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]])

    path = "./test.safetensors"
    ak.to_safetensors(array, path)

    loaded = ak.from_safetensors(path)
    virtual_loaded = ak.from_safetensors(path, virtual=True)

    os.remove(path)

    assert array.layout.is_equal_to(loaded.layout, all_parameters=True)
    assert array.layout.is_equal_to(
        virtual_loaded.layout.materialize(), all_parameters=True
    )


def test_virtual_array_to_safetensors():
    import awkward as ak

    array = ak.Array([[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]])

    path = "./test_virtual{}.safetensors".format

    ak.to_safetensors(array, path(0))
    virtual_loaded = ak.from_safetensors(path(0), virtual=True)

    ak.to_safetensors(virtual_loaded, path(1))
    loaded = ak.from_safetensors(path(1), virtual=False)

    os.remove(path(0))
    os.remove(path(1))

    assert virtual_loaded.layout.materialize().is_equal_to(
        loaded.layout, all_parameters=True
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest  # noqa: F401

import awkward as ak

ATTRS = {"foo": "bar", "@foo": "baz"}


def test_array_slice():
    array = ak.Array([[0, 1, 2], [4]], attrs=ATTRS)
    assert array.attrs == ATTRS

    assert array[0].attrs == ATTRS
    assert array[1:].attrs == ATTRS


def test_array_field():
    array = ak.Array([[{"x": 1}, {"x": 2}], [{"x": 10}]], attrs=ATTRS)
    assert array.attrs == ATTRS

    assert array.x.attrs == ATTRS
    assert array.x[1:].attrs == ATTRS


def test_record_field():
    array = ak.Array([{"x": [1, 2, 3]}], attrs=ATTRS)
    assert array.attrs == ATTRS

    record = array[0]
    assert record.attrs == ATTRS
    assert record.x.attrs == ATTRS



from __future__ import annotations

import re
from collections.abc import Iterable
from datetime import datetime

import numpy as np
import pytest

import awkward as ak
from awkward.types.numpytype import _primitive_to_dtype_dict, primitive_to_dtype


def test_issue() -> None:
    """Assert the GitHub issue #2377 is resolved."""
    empty = ak.Array([])  # <Array [] type='0 * unknown'>

    # No exception should be raised.
    result = ak.flatten(empty[empty])

    # The result should be an empty awkward array of the same type.
    expected_layout = ak.Array([]).layout
    assert result.layout.is_equal_to(expected_layout, all_parameters=True)


def _add_necessary_unit(dtype_name: str) -> str:
    """Completes datetime or timedelta dtype names with a unit if missing."""
    UNIT_LESS_DT_RE = re.compile(r"^(?:datetime|timedelta)\d*$")
    SAMPLE_UNIT = "15us"
    if UNIT_LESS_DT_RE.fullmatch(dtype_name):
        return f"{dtype_name}[{SAMPLE_UNIT}]"
    return dtype_name


# (dtype('bool'), dtype('int8'), dtype('M8[15us]'), ...) Supported dtypes
DTYPES = tuple(
    primitive_to_dtype(_add_necessary_unit(k)) for k in _primitive_to_dtype_dict.keys()
)


# (dtype('bool'), dtype('int8'), ...) Only bool and integer types
INDEXABLE_DTYPES = tuple(d for d in DTYPES if d.kind in ("b", "i", "u"))


AWKWARD_ARRAYS = (
    # Empty array of unknown type
    ak.Array([]),
    # Empty arrays of various dtypes
    *[ak.from_numpy(np.array([], dtype=d)) for d in DTYPES],
    # Non-empty arrays of specific types
    ak.Array([1, 2, 3]),
    ak.Array([[1, 2], [], [3]]),
    ak.Array([[[1.1, 2.2], []], [[3.3]], []]),
    ak.Array([1 + 1j, 2 + 2j, 3 + 3j]),
    ak.Array([datetime(2020, 1, 1), datetime(2021, 1, 1), datetime(2022, 1, 1)]),
    # Non-empty arrays of mixed types
    ak.Array([[1, 2], [1.1, 2.2], [1 + 1j, 2 + 2j]]),
    ak.Array(
        [
            [1, 2],
            [1.1, 2.2],
            [1 + 1j, 2 + 2j],
            [datetime(2020, 1, 1), datetime(2021, 1, 1)],
        ]
    ),
)

EMPTY_INDEXES = (
    # (),  # NOTE: An empty tuple doesn't pass the test.
    [],
    ak.Array([]),
    *[ak.from_numpy(np.array([], dtype=d)) for d in INDEXABLE_DTYPES],
    *[np.array([], dtype=d) for d in INDEXABLE_DTYPES],
)


@pytest.mark.parametrize("a", AWKWARD_ARRAYS)
@pytest.mark.parametrize("idx", EMPTY_INDEXES)
def test_empty_index(a: ak.Array, idx: Iterable) -> None:
    """Assert indexing with an empty array preserves the type."""
    result = a[idx]

    # Assert an empty array.
    assert result.to_list() == []

    # Assert the type is preserved
    # e.g., "0 * complex128", "0 * var * union[complex128, datetime64[us]]"
    expected_typestr = str(ak.types.ArrayType(a.type.content, 0))
    assert result.typestr == expected_typestr


def test_flatten_empty_array() -> None:
    """Assert empty arrays can be flattened unless its type forbids it."""

    # An empty array of unknown type can be flattened as unknown type can be
    # flattened, e.g., an empty array of nested arrays.
    a = ak.Array([])  #  type='0 * unknown'
    assert ak.flatten(a).to_list() == []

    # The empty sub-array is an empty array of arrays of integers.
    a = ak.Array([[[1]], []])  # type='2 * var * var * int64'
    assert ak.flatten(a, axis=2).to_list() == [[1], []]

    # The type forbids flattening.
    a = ak.from_numpy(np.array([], dtype=int))  #  type='0 * int64'
    with pytest.raises(ak.errors.AxisError):
        ak.flatten(a)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import sys

import numpy as np

import awkward as ak


def test_all_empty():
    result = ak.concatenate([ak.Array([None]) for _ in range(5000)], axis=0)
    expected = ak.Array([None] * 5000)
    assert ak.array_equal(result, expected)

    result = ak.concatenate([ak.Array([{"x": None}]) for _ in range(5000)], axis=0)
    expected = ak.Array([{"x": None}] * 5000)
    assert ak.array_equal(result, expected)

    result = ak.concatenate(
        [ak.Array([{"x": i, "y": None}]) for i in range(5000)], axis=0
    )
    expected = ak.Array([{"x": i, "y": None} for i in range(5000)])
    assert ak.array_equal(result, expected)


def test_empty_and_nonempty():
    N = sys.getrecursionlimit()
    rng = np.random.default_rng(42)
    choices1 = np.concatenate(
        (np.array([True, True]), rng.choice([True, False], size=N))
    )
    choices2 = np.concatenate(
        (np.array([False, False]), rng.choice([True, False], size=N))
    )
    choices3 = np.concatenate(
        (np.array([True, False]), rng.choice([True, False], size=N))
    )
    choices4 = np.concatenate(
        (np.array([False, True]), rng.choice([True, False], size=N))
    )
    all_choices = [choices1, choices2, choices3, choices4]

    for choices in all_choices:
        rows = []
        for i in range(N):
            if choices[i]:
                rows.append(None)
            else:
                rows.append(i)
        result = ak.concatenate([ak.Array([row]) for row in rows], axis=0)
        expected = ak.Array(rows)
        assert ak.array_equal(result, expected)

        rows = []
        for i in range(N):
            if choices[i]:
                rows.append({"x": None})
            else:
                rows.append({"x": i})
        result = ak.concatenate([ak.Array([row]) for row in rows], axis=0)
        expected = ak.Array(rows)
        assert ak.array_equal(result, expected)

        rows = []
        for i in range(N):
            if choices[i]:
                rows.append({"x": i, "y": None})
            else:
                rows.append({"x": i, "y": i})
        result = ak.concatenate([ak.Array([row]) for row in rows], axis=0)
        expected = ak.Array(rows)
        assert ak.array_equal(result, expected)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
from __future__ import annotations

import numpy as np

import awkward as ak


def test():
    """Check that to_list() does not break when the array is built from buffers
    with a length of type np.int64.
    """
    awk = ak.Array(np.ones((7, 0)))
    form, length, container = ak.to_buffers(ak.to_packed(awk))
    awk_from_buf = ak.from_buffers(form, np.int64(length), container)
    lst = awk_from_buf.to_list()

    assert len(lst) == length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    record = ak.Record({"0": {"0": 0}, "1": 0})
    assert record.to_list() == {"0": {"0": 0}, "1": 0}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    arr = ak.Array([1])
    arr.attrs["foo"] = "bar"

    arr2 = ak.copy(arr)
    assert arr2.attrs == arr.attrs

    arr2.attrs["foo"] = "baz"
    assert arr2.attrs != arr.attrs



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_read_negative_number_of_items():
    vm = ak.forth.ForthMachine32("input source -5 source #q-> stack")
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.stack == []

    vm = ak.forth.ForthMachine32("input source output sink float64 -5 source #q-> sink")
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.output("sink").tolist() == []


def test_read_negative_and_positive_number_of_items():
    vm = ak.forth.ForthMachine32(
        "input source -5 source #q-> stack 5 source #q-> stack"
    )
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.stack == [1, 2, 3, 4, 5]

    vm = ak.forth.ForthMachine32(
        "input source output sink float64 -5 source #q-> sink 5 source #q-> sink"
    )
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.output("sink").tolist() == [1, 2, 3, 4, 5]


def test_read_positive_and_negative_number_of_items():
    vm = ak.forth.ForthMachine32(
        "input source 5 source #q-> stack -5 source #q-> stack"
    )
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.stack == [1, 2, 3, 4, 5]

    vm = ak.forth.ForthMachine32(
        "input source output sink float64 5 source #q-> sink -5 source #q-> sink"
    )
    vm.run({"source": np.array([1, 2, 3, 4, 5], dtype=np.int64)})
    assert vm.output("sink").tolist() == [1, 2, 3, 4, 5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.shape import unknown_length
from awkward._nplikes.virtual import VirtualNDArray


@pytest.mark.parametrize("offsets_length", [5, unknown_length])
@pytest.mark.parametrize("content_length", [9, unknown_length])
def test(offsets_length, content_length):
    offset_generator = lambda: np.array([0, 2, 4, 5, 6], dtype=np.int64)  # noqa: E731
    data_generator = lambda: np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.int64)  # noqa: E731

    offsets = ak.index.Index(
        VirtualNDArray(
            Numpy.instance(),
            shape=(offsets_length,),
            dtype=np.int64,
            generator=offset_generator,
        )
    )
    data = ak.contents.NumpyArray(
        VirtualNDArray(
            Numpy.instance(),
            shape=(content_length,),
            dtype=np.int64,
            generator=data_generator,
        )
    )
    array = ak.Array(ak.contents.ListOffsetArray(offsets, data))
    assert array.to_list() == [[1, 2], [3, 4], [5], [6]]

    buffers = {"node0-offsets": offset_generator, "node1-data": data_generator}
    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("int64", form_key="node1"), form_key="node0"
    )
    array = ak.from_buffers(form, 4, buffers)
    assert array.to_list() == [[1, 2], [3, 4], [5], [6]]

    offsets = VirtualNDArray(
        Numpy.instance(),
        shape=(offsets_length,),
        dtype=np.int64,
        generator=offset_generator,
    )
    data = VirtualNDArray(
        Numpy.instance(),
        shape=(content_length,),
        dtype=np.int64,
        generator=data_generator,
    )
    buffers = {"node0-offsets": offsets, "node1-data": data}
    form = ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("int64", form_key="node1"), form_key="node0"
    )
    array = ak.from_buffers(form, 4, buffers)
    assert array.to_list() == [[1, 2], [3, 4], [5], [6]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

pa = pytest.importorskip("pyarrow")


def test():
    one = ak.zip({"pt": [[1, 2], [1]], "eta": [[1.1, 2.1], [1]]})
    two = ak.zip({"pt": [[2.1], [2.21, 2.22]], "eta": [[2.11], [2.221, 2.222]]})
    assert ak.cartesian({"foo": one, "bar": two}, nested=False).layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 2, 4]),
            ak.contents.RecordArray(
                [
                    ak.contents.IndexedArray(
                        ak.index.Index64([0, 1, 2, 2]),
                        ak.contents.RecordArray(
                            [
                                ak.contents.NumpyArray(
                                    np.array([1, 2, 1], dtype=np.int64)
                                ),
                                ak.contents.NumpyArray(
                                    np.array([1.1, 2.1, 1.0], dtype=np.float64)
                                ),
                            ],
                            ["pt", "eta"],
                        ),
                    ),
                    ak.contents.IndexedArray(
                        ak.index.Index64([0, 0, 1, 2]),
                        ak.contents.RecordArray(
                            [
                                ak.contents.NumpyArray(
                                    np.array([2.1, 2.21, 2.22], dtype=np.float64)
                                ),
                                ak.contents.NumpyArray(
                                    np.array([2.11, 2.221, 2.222], dtype=np.float64)
                                ),
                            ],
                            ["pt", "eta"],
                        ),
                    ),
                ],
                ["foo", "bar"],
            ),
        )
    )
    assert ak.cartesian({"foo": one, "bar": two}, nested=True).layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 2, 3]),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 1, 2, 4]),
                ak.contents.RecordArray(
                    [
                        ak.contents.IndexedArray(
                            ak.index.Index64([0, 1, 2, 2]),
                            ak.contents.RecordArray(
                                [
                                    ak.contents.NumpyArray(
                                        np.array([1, 2, 1], dtype=np.int64)
                                    ),
                                    ak.contents.NumpyArray(
                                        np.array([1.1, 2.1, 1.0], dtype=np.float64)
                                    ),
                                ],
                                ["pt", "eta"],
                            ),
                        ),
                        ak.contents.IndexedArray(
                            ak.index.Index64([0, 0, 1, 2]),
                            ak.contents.RecordArray(
                                [
                                    ak.contents.NumpyArray(
                                        np.array([2.1, 2.21, 2.22], dtype=np.float64)
                                    ),
                                    ak.contents.NumpyArray(
                                        np.array([2.11, 2.221, 2.222], dtype=np.float64)
                                    ),
                                ],
                                ["pt", "eta"],
                            ),
                        ),
                    ],
                    ["foo", "bar"],
                ),
            ),
        )
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test_simple_field_replacement_preserves_order():
    """Test that replacing an existing field preserves its position in the field order."""
    array = ak.Array([{"x": 1, "y": 2, "z": 3}])

    # Replace middle field
    result = ak.with_field(array, [100], "y")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [{"x": 1, "y": 100, "z": 3}]

    # Replace first field
    result = ak.with_field(array, [100], "x")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [{"x": 100, "y": 2, "z": 3}]

    # Replace last field
    result = ak.with_field(array, [100], "z")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [{"x": 1, "y": 2, "z": 100}]


def test_simple_field_addition_appends():
    """Test that adding a new field appends it to the end."""
    array = ak.Array([{"x": 1, "y": 2, "z": 3}])

    result = ak.with_field(array, [100], "w")
    assert ak.fields(result) == ["x", "y", "z", "w"]
    assert result.to_list() == [{"x": 1, "y": 2, "z": 3, "w": 100}]


def test_setitem_replacement_preserves_order():
    """Test that using array['field'] = value preserves field order."""
    array = ak.Array([{"x": 1, "y": 2, "z": 3}])

    # Replace middle field using setitem
    array["y"] = [100]
    assert ak.fields(array) == ["x", "y", "z"]
    assert array.to_list() == [{"x": 1, "y": 100, "z": 3}]


def test_setitem_addition_appends():
    """Test that using array['new_field'] = value appends the field."""
    array = ak.Array([{"x": 1, "y": 2, "z": 3}])

    array["w"] = [100]
    assert ak.fields(array) == ["x", "y", "z", "w"]
    assert array.to_list() == [{"x": 1, "y": 2, "z": 3, "w": 100}]


def test_nested_field_replacement_preserves_order():
    """Test that replacing a nested field preserves order at both levels."""
    array = ak.Array([{"a": {"x": 1, "y": 2, "z": 3}, "b": 10, "c": 20}])

    # Replace nested field "y" inside "a"
    result = ak.with_field(array, [100], ["a", "y"])
    assert ak.fields(result) == ["a", "b", "c"]
    assert ak.fields(result["a"]) == ["x", "y", "z"]
    assert result.to_list() == [{"a": {"x": 1, "y": 100, "z": 3}, "b": 10, "c": 20}]


def test_nested_field_replacement_with_setitem():
    """Test nested field replacement using setitem syntax."""
    array = ak.Array([{"a": {"x": 1, "y": 2, "z": 3}, "b": 10, "c": 20}])

    # Replace nested field using setitem
    array["a", "y"] = [100]
    assert ak.fields(array) == ["a", "b", "c"]
    assert ak.fields(array["a"]) == ["x", "y", "z"]
    assert array.to_list() == [{"a": {"x": 1, "y": 100, "z": 3}, "b": 10, "c": 20}]


def test_nested_field_addition():
    """Test that adding a new nested field appends it."""
    array = ak.Array([{"a": {"x": 1, "y": 2, "z": 3}, "b": 10, "c": 20}])

    # Add new nested field
    result = ak.with_field(array, [100], ["a", "w"])
    assert ak.fields(result) == ["a", "b", "c"]
    assert ak.fields(result["a"]) == ["x", "y", "z", "w"]
    assert result.to_list() == [
        {"a": {"x": 1, "y": 2, "z": 3, "w": 100}, "b": 10, "c": 20}
    ]


def test_multiple_replacements_preserve_order():
    """Test that multiple sequential replacements preserve order."""
    array = ak.Array([{"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}])

    # Replace multiple fields in different orders
    result = array
    result = ak.with_field(result, [20], "b")
    result = ak.with_field(result, [40], "d")
    result = ak.with_field(result, [10], "a")

    assert ak.fields(result) == ["a", "b", "c", "d", "e"]
    assert result.to_list() == [{"a": 10, "b": 20, "c": 3, "d": 40, "e": 5}]


def test_replacement_with_array_data():
    """Test field replacement with various array structures."""
    array = ak.Array(
        [
            {"x": 1, "y": [1, 2], "z": 3},
            {"x": 2, "y": [3, 4, 5], "z": 4},
        ]
    )

    # Replace field with different array structure
    result = ak.with_field(array, [[10, 20], [30, 40, 50]], "y")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [
        {"x": 1, "y": [10, 20], "z": 3},
        {"x": 2, "y": [30, 40, 50], "z": 4},
    ]


def test_replacement_with_scalar_broadcast():
    """Test field replacement with scalar broadcasting."""
    array = ak.Array(
        [
            {"x": 1, "y": 2, "z": 3},
            {"x": 4, "y": 5, "z": 6},
            {"x": 7, "y": 8, "z": 9},
        ]
    )

    # Replace with scalar (should broadcast)
    result = ak.with_field(array, 100, "y")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [
        {"x": 1, "y": 100, "z": 3},
        {"x": 4, "y": 100, "z": 6},
        {"x": 7, "y": 100, "z": 9},
    ]


def test_tuple_field_replacement():
    """Test replacing a field in a tuple (which converts it to a record)."""
    array = ak.Array([(1, 2, 3)])

    # Replacing field "1" should preserve position
    result = ak.with_field(array, [100], "1")
    assert ak.fields(result) == ["0", "1", "2"]
    assert result.to_list() == [{"0": 1, "1": 100, "2": 3}]


def test_single_field_replacement():
    """Test replacing the only field in a record."""
    array = ak.Array([{"x": 1}, {"x": 2}, {"x": 3}])

    result = ak.with_field(array, [100, 200, 300], "x")
    assert ak.fields(result) == ["x"]
    assert result.to_list() == [{"x": 100}, {"x": 200}, {"x": 300}]


def test_deeply_nested_replacement():
    """Test replacing a deeply nested field."""
    array = ak.Array([{"a": {"b": {"x": 1, "y": 2, "z": 3}, "c": 10}, "d": 100}])

    # Replace deeply nested field
    result = ak.with_field(array, [999], ["a", "b", "y"])
    assert ak.fields(result) == ["a", "d"]
    assert ak.fields(result["a"]) == ["b", "c"]
    assert ak.fields(result["a", "b"]) == ["x", "y", "z"]
    assert result.to_list() == [
        {"a": {"b": {"x": 1, "y": 999, "z": 3}, "c": 10}, "d": 100}
    ]


def test_where_none_always_appends():
    """Test that where=None always appends a new field."""
    array = ak.Array([{"x": 1, "y": 2}])

    # where=None should always append
    result = ak.with_field(array, [100], where=None)
    assert ak.fields(result) == ["x", "y", "2"]
    assert result.to_list() == [{"x": 1, "y": 2, "2": 100}]

    # Multiple where=None should keep appending
    result = ak.with_field(result, [200], where=None)
    assert ak.fields(result) == ["x", "y", "2", "3"]
    assert result.to_list() == [{"x": 1, "y": 2, "2": 100, "3": 200}]


def test_mixed_operations():
    """Test a mix of additions and replacements."""
    array = ak.Array([{"x": 1, "y": 2, "z": 3}])

    # Add a new field
    result = ak.with_field(array, [10], "a")
    assert ak.fields(result) == ["x", "y", "z", "a"]

    # Replace an existing field
    result = ak.with_field(result, [20], "y")
    assert ak.fields(result) == ["x", "y", "z", "a"]

    # Add another new field
    result = ak.with_field(result, [30], "b")
    assert ak.fields(result) == ["x", "y", "z", "a", "b"]

    # Replace the first field
    result = ak.with_field(result, [100], "x")
    assert ak.fields(result) == ["x", "y", "z", "a", "b"]

    assert result.to_list() == [{"x": 100, "y": 20, "z": 3, "a": 10, "b": 30}]


def test_replacement_in_list_of_records():
    """Test field replacement in a list of records."""
    array = ak.Array(
        [
            [{"x": 1, "y": 2, "z": 3}, {"x": 4, "y": 5, "z": 6}],
            [{"x": 7, "y": 8, "z": 9}],
        ]
    )

    result = ak.with_field(array, [[100, 200], [300]], "y")
    assert ak.fields(result) == ["x", "y", "z"]
    assert result.to_list() == [
        [{"x": 1, "y": 100, "z": 3}, {"x": 4, "y": 200, "z": 6}],
        [{"x": 7, "y": 300, "z": 9}],
    ]
